/*************************************************  DEFAULT OPTIONS ***************************************************/

let oldLocationChangePayload = false;
let routerReducerKey = 'router';
let isReduxTravelling = false;

/*************************************************  REDUX ACTIONS *****************************************************/

const CALL_HISTORY_METHOD = '@@router/CALL_HISTORY_METHOD';
export const LOCATION_CHANGE = '@@router/LOCATION_CHANGE';

const locationChangeAction = (location, action) => ({
  type: LOCATION_CHANGE,
  payload: oldLocationChangePayload ? { ...location, action } : { location, action },
});

function updateLocation(method) {
  return (...args) => ({
    type: CALL_HISTORY_METHOD,
    payload: { method, args },
  });
}

export const push = updateLocation('push');
export const replace = updateLocation('replace');
export const go = updateLocation('go');
export const goBack = updateLocation('goBack');
export const goForward = updateLocation('goForward');

/*************************************************  REDUX REDUCER *****************************************************/

const initialState = {
  location: null,
  action: null,
};

export const routerReducer = (state = initialState, { type, payload } = {}) => {
  if (type === LOCATION_CHANGE) {
    if (oldLocationChangePayload) {
      const { action, ...location } = payload || {};
      return { ...state, location, action };
    }
    const { location, action } = payload || {};
    return { ...state, location, action };
  }
  return state;
};

/*************************************************  REDUX MIDDLEWARE **************************************************/

// eslint-disable-next-line
export const routerMiddleware = history => () => next => (action) => {
  if (action.type !== CALL_HISTORY_METHOD) {
    return next(action);
  }
  const { payload: { method, args } } = action;
  history[method](...args);
};

/**********************************************  REDUX TRAVELLING  ***************************************************/

const handleReduxTravelling = (store, history) => {
  const locationEqual = (loc1, loc2) => (loc1.pathname === loc2.pathname && loc1.search === loc2.search && loc1.hash === loc2.hash);

  return store.subscribe(() => {
    const sLoc = store.getState()[routerReducerKey].location;
    const hLoc = history.location;

    if (!locationEqual(sLoc, hLoc)) {
      isReduxTravelling = true;
      history.push({ pathname: sLoc.pathname, search: sLoc.search, hash: sLoc.hash });
    }
  });
};

/******************************************  REDUX FIRST HISTORY   *****************************************************/

export const createReduxHistory = (store, history, options = {}) => {
  //set options
  oldLocationChangePayload = !!options.oldLocationChangePayload;
  routerReducerKey = options.routerReducerKey || routerReducerKey;

  if (options.reduxTravelling) {
    handleReduxTravelling(store, history);
  }

  let registeredCallback = [];

  //init location store
  store.dispatch(locationChangeAction(history.location, history.action));

  //listen to history API
  history.listen((location, action) => {
    if (isReduxTravelling) {
      isReduxTravelling = false;
      return;
    }
    store.dispatch(locationChangeAction(location, action));

    const state = store.getState();
    registeredCallback.forEach(c => c(state[routerReducerKey].location, state[routerReducerKey].action));
  });

  const reduxFirstHistory = {
    createHref: history.createHref,
    push: location => store.dispatch(push(location)),
    replace: location => store.dispatch(replace(location)),
    go: location => store.dispatch(go(location)),
    goBack: location => store.dispatch(goBack(location)),
    goForward: location => store.dispatch(goForward(location)),

    //listen tunnel
    listen: (callback) => {
      if (registeredCallback.indexOf(callback) < 0) {
        registeredCallback.push(callback);
      }
      return () => {
        registeredCallback = registeredCallback.filter(c => c !== callback);
      };
    },
  };

    //location tunnel
  Object.defineProperty(reduxFirstHistory, 'location', {
    get() {
      return store.getState()[routerReducerKey].location;
    },
  });

  //location tunnel
  Object.defineProperty(reduxFirstHistory, 'action', {
    get() {
      return store.getState()[routerReducerKey].action;
    },
  });

  return reduxFirstHistory;
};
